from pymodaq.daq_move.utility_classes import DAQ_Move_base
from pymodaq.daq_move.utility_classes import comon_parameters
from pymodaq.daq_utils.daq_utils import ThreadCommand
from easydict import EasyDict as edict
from ..hardware.smaract.smaract import SmarAct
from ..hardware.smaract.smaract import get_controller_locators


class DAQ_Move_SmarActMCS(DAQ_Move_base):
    """
    This plugin supports only SmarAct LINEAR positionners (SLC type), with
    enabled sensors attached to them.
    We suppose to have one (or multiple) MCS controllers connected. With 3
    channels (each).
    We suppose that the configuration of the controllers (sensor type etc) has
    been done via the SmarAct MCS Configuration software.
    Tested with one SLC-1740-S (closed loop with nanometer precision sensor)
    connected via a MCS-3S-EP-SDS15-TAB (sensor module) to a MCS-3D (or MCS-3C)
    controller on Windows 7.
    """
    _controller_units = "µm"

    # find controller locators
    controller_locators = get_controller_locators()

    is_multiaxes = True
    # we suppose to have a MCS controller with 3 channels (like the MCS-3D).
    stage_names = [0, 1, 2]
    # bounds corresponding to the SLC-24180
    min_bound = -61500  # µm
    max_bound = +61500  # µm
    offset = 0  # µm

    params = [
        {
            "title": "group parameter:",
            "name": "group_parameter",
            "type": "group",
            "children": [
                {
                    "title": "Controller Name:",
                    "name": "smaract_mcs",
                    "type": "str",
                    "value": "SmarAct MCS controller",
                    "readonly": True,
                },
                {
                    "title": "Controller locator",
                    "name": "controller_locator",
                    "type": "list",
                    "limits": controller_locators,
                },
            ],
        },
        ##########################################################
        # the ones below should ALWAYS be present!!!
        {
            "title": "MultiAxes:",
            "name": "multiaxes",
            "type": "group",
            "visible": is_multiaxes,
            "children": [
                {
                    "title": "is Multiaxes:",
                    "name": "ismultiaxes",
                    "type": "bool",
                    "value": is_multiaxes,
                    "default": False,
                },
                {
                    "title": "Status:",
                    "name": "multi_status",
                    "type": "list",
                    "value": "Master",
                    "limits": ["Master", "Slave"],
                },
                {
                    "title": "Axis:",
                    "name": "axis",
                    "type": "list",
                    "limits": stage_names,
                },
            ],
        },
    ] + comon_parameters
    ##########################################################

    def __init__(self, parent=None, params_state=None):
        super().__init__(parent, params_state)

        self.controller = None
        self.settings.child("epsilon").setValue(0.002)

    def ini_stage(self, controller=None):
        """Initialize the controller and stages (axes) with given parameters.

        Parameters
        ----------
        controller (object): custom object of a PyMoDAQ plugin (Slave case).
            None if only one actuator by controller (Master case)

        Returns
        -------
        self.status: (edict) with initialization status: three fields:
            * info: (str)
            * controller: (object) initialized controller
            * initialized: (bool) False if initialization failed otherwise
                True
        """
        try:
            # initialize the stage and its controller status
            # controller is an object that may be passed to other instances of
            # DAQ_Move_Mock in case of one controller controlling
            # multiactuators (or detector)
            self.status.update(edict(
                info="", controller=None, initialized=False))

            # check whether this stage is controlled by a multiaxe controller
            # (to be defined for each plugin)
            # if multiaxes then init the controller here if Master state
            # otherwise use external controller
            if self.settings.child('multiaxes',
                                   'ismultiaxes').value() \
                    and self.settings.child('multiaxes',
                                            'multi_status').value() == "Slave":
                if controller is None:
                    raise Exception('No controller has been defined externally'
                                    ' while this axe is a slave one')
                else:
                    self.controller = controller
            else:  # Master stage
                self.controller = SmarAct()
                self.controller.init_communication(
                    self.settings.child("group_parameter",
                                        "controller_locator").value()
                )

            # min and max bounds will depend on which positionner is plugged.
            # Anyway the bounds are secured by the library functions.
            self.settings.child("bounds", "is_bounds").setValue(True)
            self.settings.child("bounds", "min_bound").setValue(self.min_bound)
            self.settings.child("bounds", "max_bound").setValue(self.max_bound)
            self.settings.child("scaling", "use_scaling").setValue(True)
            self.settings.child("scaling", "offset").setValue(self.offset)

            self.status.controller = self.controller
            self.status.initialized = True

            return self.status

        except Exception as e:
            self.emit_status(ThreadCommand("Update_Status", [str(e), "log"]))
            self.status.info = str(e)
            self.status.initialized = False

            return self.status

    def close(self):
        """Close the communication with the SmarAct controller.
        """

        self.controller.close_communication()
        self.controller = None

    def check_position(self):
        """Get the current position from the hardware with scaling conversion.

        Returns
        -------
        float: The position obtained after scaling conversion.
        """
        position = self.controller.get_position(
            self.settings.child("multiaxes", "axis").value()
        )

        # the position given by the controller is in nanometers, we convert in
        # micrometers
        position = float(position) / 1e3

        # convert position if scaling options have been used, mandatory here
        position = self.get_position_with_scaling(position)
        self.current_position = position
        self.emit_status(ThreadCommand("check_position", [position]))

        return position

    def move_Abs(self, position):
        """Move to an absolute position

        Parameters
        ----------
        position: float
        """
        # limit position if bounds options has been selected and if position is
        # out of them
        position = self.check_bound(position)
        self.target_position = position
        # convert the user set position to the controller position if scaling
        # has been activated by user
        position = self.set_position_with_scaling(position)

        # we convert position in nm
        position = int(position * 1e3)

        # the SmarAct controller asks for nanometers
        self.controller.absolute_move(
            self.settings.child("multiaxes", "axis").value(), position
        )

        # start polling the position until the actuator reach the target
        # position within epsilon
        # defined as a parameter field (comon_parameters)
        self.poll_moving()

    def move_Rel(self, position):
        """Move to a relative position

        Parameters
        ----------
        position: float
        """
        # limit position if bounds options has been selected and if position is
        # out of them
        position = (
            self.check_bound(self.current_position + position)
            - self.current_position)
        self.target_position = position + self.current_position
        # convert the user set position to the controller position if scaling
        # has been activated by user
        position = self.set_position_with_scaling(position)

        # we convert position in nm
        position = int(position * 1e3)

        # the SmarAct controller asks for nanometers
        self.controller.relative_move(
            self.settings.child("multiaxes", "axis").value(), position
        )

        self.poll_moving()

    def move_Home(self):
        """Move to home and reset position to zero.
        """

        self.controller.find_reference(
            self.settings.child("multiaxes", "axis").value())

    def stop_motion(self):
        """
        See Also
        --------
        DAQ_Move_base.move_done
        """

        self.controller.stop(self.settings.child("multiaxes", "axis").value())

        self.move_done()


if __name__ == "__main__":
    test = DAQ_Move_SmarActMCS()



# -*- coding: utf-8 -*-

import ctypes
import os
import re

"""The API documentation is in SmarAct MCS2 Programmers Guide, which should be
   in the same folder as this file.
    
    Prerequisites
    -------------
    We suppose that the configuration of the channels of the controller (sensor
type etc) has been done via the SmarAct MCS2ServiceTool software.
    We suppose to have some linear positionners (SLC type) with sensors
(S option).

    Tested with:
    - positioner: SLC24180s
    - sensor type: MCS2-S-0001
    - controller: MCS2 in closed loop, positioner type = SL...S1SS (300)
"""

# The SmarActCTL.dll should also be in the current folder, but also
# SmarActCTL.lib, SmarActIO.dll, SmarActSI.dll, SmarActSI.lib
# The CDLL function asks for the full path
dir_path = os.path.dirname(os.path.realpath(__file__))
smaract_dll = ctypes.CDLL(os.path.join(dir_path, "SmarActCTL.dll"))


def get_controller_locators():
    """Get the controller locator of the plugged MCS2 controllers.

    e.g. usb:sn:00000667
    The 8-digit number corresponds to the serial number printed on the device.

    Returns
    -------
    controller_locators : list of str
    """
    ioListSize = 4096
    options = ctypes.c_char()
    outList = (' ' * ioListSize).encode()
    ioListSize = ctypes.c_ulong(ioListSize)

    status = smaract_dll.SA_CTL_FindDevices(
        ctypes.byref(options),
        outList,
        ctypes.byref(ioListSize)
    )

    if status != 0:
        raise Exception('SmarAct SA_CTL_FindDevices error')

    controller_locators = re.findall("usb:sn:MCS2-[0-9]*", outList.decode())

    if not controller_locators:
        raise Exception('SmarActMCS2Wrapper: No controller found !')

    return controller_locators


class SmarActMCS2Wrapper:

    def __init__(self):
        super(SmarActMCS2Wrapper, self).__init__()

        # After initialization the program gives an unsigned int index to each
        # controller that is connected.
        # The plugin does not support multiple controllers management.
        self.controller_index = ''

    def init_communication(self, controller_locator):
        """Use the controller locator returned from get_controller_locator and
            return the system index attributed by the dll to refer to the
            controller.

        Parameters
        -------
        controller_locator: str
        """
        controller_index = ctypes.c_ulong()
        # we choose the synchronous communication mode
        # options = 'sync'.encode('ascii')

        status = smaract_dll.SA_CTL_Open(
            ctypes.byref(controller_index),
            controller_locator.encode('ascii'),
            ""
        )

        if status != 0:
            raise Exception('SmarActMCS2Wrapper: SmarAct SA_CTL_Open failed')

        self.controller_index = controller_index.value

    def get_number_of_channels(self):
        """
            Return the number of channels of the controller. Note that the
            number of channels does not represent the number positioners and/or
            end effectors that are currently connected to the system.

        Returns
        -------
        numberOfChannels.value: unsigned int
        """
        number_of_channels = ctypes.c_ulong()

        # The code 0x020F0017 corresponds to SA_CTL_PKEY_NUMBER_OF_CHANNELS has
        # been found in page 137 of the programmers guide
        status = smaract_dll.SA_CTL_GetProperty_i32(
            ctypes.c_ulong(self.controller_index),
            0,
            0x020F0017,
            ctypes.byref(number_of_channels),
            0)

        if status != 0:
            self.close_communication()
            raise Exception(
                'SmarActMCS2Wrapper: SmarAct API failed to get the number of'
                'channels')

        return number_of_channels.value

    def close_communication(self):
        """Close the communication with the controller.
        """
        status = smaract_dll.SA_CTL_Close(
            ctypes.c_ulong(self.controller_index)
        )

        if status != 0:
            raise Exception('SmarActMCS2Wrapper: SmarAct SA_CTL_Close failed')

    def get_position(self, channel_index):
        """Return the current position of the positioner connected to the
            channel indexed by channel_index (starts at 0) in picometers.

        Parameters
        ----------
        channel_index: unsigned int

        Returns
        -------
        position.value: signed int
        """

        position = ctypes.c_longlong()

        # The code 0x0305001D corresponds to SA_CTL_PKEY_POSITION has been
        # found in page 155 of the programmers guide
        status = smaract_dll.SA_CTL_GetProperty_i64(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            0x0305001D,
            ctypes.byref(position),
            0
        )

        if status != 0:
            self.close_communication()
            raise Exception('SmarActMCS2Wrapper: SmarAct API failed to get the'
                            ' position of the actuator')

        return position.value

    def find_reference(self, channel_index):
        """Find the physical zero reference of the positioner (starting in the
            forward direction) and reset the position to zero.
            Here we suppose that the reference method has been configured with
            the MCS Configuration software for example.

        Parameters
        ----------
        channel_index: unsigned int
        """

        status = smaract_dll.SA_CTL_Reference(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            0
        )

        if status != 0:
            self.close_communication()
            raise Exception(
                'SmarActMCS2Wrapper: SmarAct SA_CTL_Reference failed')

    def relative_move(self, channel_index, relative_move_value):
        """Execute a relative move in picometers.
            If a mechanical end stop is detected while the command is in
            execution, the movement will be aborted (without notice).

        Parameters
        ----------
        channel_index: unsigned int
        relative_move_value: signed int. Relative distance in picometer.
        """

        # To perform a relative movement it is necessary to configure the move
        # mode before.
        # The code 0x03050087 corresponds to SA_CTL_PKEY_MOVE_MODE has been
        # found in page 152 of the programmers guide.
        # Number 1 corresponds to a relative move (4th parameter). Found in
        # page 153 of the programmers guide.
        status = smaract_dll.SA_CTL_SetProperty_i32(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            0x03050087,
            1
        )

        if status != 0:
            self.close_communication()
            raise Exception(
                'SmarActMCS2Wrapper: SmarAct API failed to configure the'
                'relative move mode.'
            )

        status = smaract_dll.SA_CTL_Move(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            ctypes.c_longlong(relative_move_value),
            0
        )

        if status != 0:
            self.close_communication()
            raise Exception('SmarActMCS2Wrapper: SA_CTL_Move failed.')

    def absolute_move(self, channel_index, absolute_move_value):
        """Go to an absolute position in picometers.
            If a mechanical end stop is detected while the command is in
            execution, the movement will be aborted (without notice).

        Parameters
        ----------
        channel_index: unsigned int
        absolute_move_value: signed int. Absolute position in picometer.
        """

        # To perform an absolute movement it is necessary to configure the move
        # mode before.
        # The code 0x03050087 corresponds to SA_CTL_PKEY_MOVE_MODE has been
        # found in page 152 of the programmers guide.
        # Number 0 corresponds to an absolute move (4th parameter). Found in
        # page 153 of the programmers guide.
        status = smaract_dll.SA_CTL_SetProperty_i32(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            0x03050087,
            0
        )

        if status != 0:
            self.close_communication()
            raise Exception(
                'SmarActMCS2Wrapper: SmarAct API failed to configure the'
                'absolute move mode.'
            )

        status = smaract_dll.SA_CTL_Move(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            ctypes.c_longlong(absolute_move_value),
            0
        )

        if status != 0:
            self.close_communication()
            raise Exception(
                'SmarActMCS2Wrapper: SA_CTL_Move failed.')

    def stop(self, channel_index):
        """Stop any ongoing movement of the positionner. This command also
            stops the hold position feature of closed-loop commands.

        Parameters
        ----------
        channel_index: unsigned int
        """

        status = smaract_dll.SA_CTL_Stop(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            0
        )

        if status != 0:
            self.close_communication()
            raise Exception('SmarActMCS2Wrapper: SmarAct SA_CTL_Stop failed')




# -*- coding: utf-8 -*-

import ctypes
import os
import re

"""
The documentation of the .dll is in SmarAct MCS Programmers Guide.
We suppose that the configuration of the controllers (sensor type etc) has been done via the SmarAct MCS Configuration
software.
We suppose to have some linear positionners (SLC type) with enabled sensors attached to them, connected to the channel 0
of the controller.
Tested with SLC-1740-S (closed loop with nanometer precision sensor) connected to a MCS-3D or MCS-3C controller.
"""

# We suppose the .dll library is in the same directory
# The CDLL function asks for the full path
dir_path = os.path.dirname(os.path.realpath(__file__))
SmaractDll = ctypes.CDLL(os.path.join(dir_path, "MCSControl.dll"))


def get_controller_locators():
    """
        Get the locators (e.g. usb:id:3118167233) of the plugged MCS controllers.

    Returns
    -------
    controller_locators : list of str
    """
    ioListSize = 4096
    options = ctypes.c_char()
    outList = (' ' * ioListSize).encode()
    ioListSize = ctypes.c_ulong(ioListSize)

    status = SmaractDll.SA_FindSystems(
        ctypes.byref(options),
        outList,
        ctypes.byref(ioListSize)
    )

    if status != 0:
        raise Exception('SmarAct SA_FindSystems error')

    controller_locators = re.findall("usb:id:[0-9]*", outList.decode())

    if not controller_locators:
        raise Exception('No controller found')

    return controller_locators


class SmarAct(object):

    def __init__(self):
        super(SmarAct, self).__init__()

        self.controller_index = ''

    def init_communication(self, controller_locator):
        """
            Use the controller locator returned from get_controller_locator and return the system index attributed by
            the dll to refer to the controller.

        Parameters
        -------
        controller_locator: str
        """
        controller_index = ctypes.c_ulong()
        # we choose the synchronous communication mode
        options = 'sync'.encode('ascii')

        status = SmaractDll.SA_OpenSystem(
            ctypes.byref(controller_index),
            controller_locator.encode('ascii'),
            options
        )

        if status != 0:
            raise Exception('SmarAct SA_OpenSystem failed')

        # return controller_index.value
        self.controller_index = controller_index.value

    def get_number_of_channels(self):
        """
            Return the number of channels of the controller. Note that the number of channels does not represent the
            number positioners and/or end effectors that are currently connected to the system.

        Parameters
        -------
        controller_index: str

        Returns
        -------
        numberOfChannels.value: unsigned int
        """
        numberOfChannels = ctypes.c_ulong()

        status = SmaractDll.SA_GetNumberOfChannels(
            ctypes.c_ulong(self.controller_index),
            ctypes.byref(numberOfChannels)
        )

        if status != 0:
            self.close_communication()
            raise Exception('SmarAct SA_GetNumberOfChannels failed')

        return numberOfChannels.value

    def close_communication(self):
        """
            Close the communication with the controller.
        """
        status = SmaractDll.SA_CloseSystem(
            ctypes.c_ulong(self.controller_index)
        )

        if status != 0:
            raise Exception(f'SmarAct SA_CloseSystem failed with error {status}')

    def get_position(self, channel_index):
        """
            Return the current position of the positioner connected to the channel indexed by channel_index
            (starts at 0) in nanometers.

        Parameters
        ----------
        channel_index: unsigned int

        Returns
        -------
        position.value: signed int
        """

        position = ctypes.c_long()

        status = SmaractDll.SA_GetPosition_S(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            ctypes.byref(position)
        )

        if status != 0:
            self.close_communication()
            raise Exception('SmarAct SA_GetPosition failed')

        return position.value

    def find_reference(self, channel_index):
        """
            Find the physical zero reference of the positioner (starting in the
            forward direction) and reset the position to zero.

        Parameters
        ----------
        channel_index: unsigned int
        """

        # with direction = 0 search for reference starts in the forward
        # direction
        direction = 0
        # hold time = 60,000 ms corresponds to infinite holding
        hold_time = 60000
        # auto zero = 1 will reset the position to zero after reaching
        # the reference mark
        auto_zero = 1

        status = SmaractDll.SA_FindReferenceMark_S(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            ctypes.c_ulong(direction),
            ctypes.c_ulong(hold_time),
            ctypes.c_ulong(auto_zero)
        )

        if status != 0:
            self.close_communication()
            raise Exception('SmarAct SA_FindReferenceMark failed')

        print('The positionner is referenced !')

    def relative_move(self, channel_index, relative_position):
        """
            Execute a relative move in nanometers
            If a mechanical end stop is detected while the command is in execution,
            the movement will be aborted (without notice).

        Parameters
        ----------
        channel_index: unsigned int
        relative_position: signed int
        """

        # hold time = 60,000 ms corresponds to infinite holding
        hold_time = 60000

        status = SmaractDll.SA_GotoPositionRelative_S(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            ctypes.c_long(relative_position),
            ctypes.c_ulong(hold_time)
        )

        if status != 0:
            self.close_communication()
            raise Exception('SmarAct SA_GotoPositionRelative failed')

    def absolute_move(self, channel_index, absolute_position):
        """
            Go to an absolute position in nanometers
            If a mechanical end stop is detected while the command is in execution,
            the movement will be aborted (without notice).

        Parameters
        ----------
        channel_index: unsigned int
        absolute_position: signed int
        """

        # hold time = 60,000 ms corresponds to infinite holding
        hold_time = 60000

        status = SmaractDll.SA_GotoPositionAbsolute_S(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index),
            ctypes.c_long(absolute_position),
            ctypes.c_ulong(hold_time)
        )

        if status != 0:
            self.close_communication()
            raise Exception('SmarAct SA_GotoPositionAbsolute failed')

    def stop(self, channel_index):
        """
            Stop any ongoing movement of the positionner. This command also stops the hold position feature of
            closed-loop commands.

        Parameters
        ----------
        channel_index: unsigned int
        """

        status = SmaractDll.SA_Stop_S(
            ctypes.c_ulong(self.controller_index),
            ctypes.c_ulong(channel_index)
        )

        if status != 0:
            self.close_communication()
            raise Exception('SmarAct SA_Stop failed')



from pymodaq.daq_move.utility_classes import DAQ_Move_base, main
from pymodaq.daq_move.utility_classes import comon_parameters
from pymodaq.daq_utils.daq_utils import ThreadCommand
from easydict import EasyDict as edict
from instrumental import instrument, Q_, list_instruments

psets = list_instruments(module='motion._smaract')

psets_str = [f"Dev. Id{pset['id']} channel {pset['index']}" for pset in psets]



class DAQ_Move_SmarAct(DAQ_Move_base):
    """

    """
    _controller_units = ""

    # find controller locators


    is_multiaxes = False
    stage_names = []

    params = [
                 {'title': 'Device', 'name': 'device', 'type': 'list', 'limits': psets_str},
                 {'title': 'Frequency (Hz)', 'name': 'frequency', 'type': 'int', 'value': 450},
                 {'title': 'Amplitude (V)', 'name': 'amplitude', 'type': 'int', 'value': 100},
                 {'title': 'Max Frequency (Hz)', 'name': 'maxfreq', 'type': 'int', 'value': 18500},

        ##########################################################
        # the ones below should ALWAYS be present!!!
        {"title": "MultiAxes:", "name": "multiaxes", "type": "group", "visible": is_multiaxes, "children": [
            {"title": "is Multiaxes:", "name": "ismultiaxes", "type": "bool", "value": is_multiaxes},
            {"title": "Status:", "name": "multi_status", "type": "list", "value": "Master",
             "limits": ["Master", "Slave"],},
            {"title": "Axis:", "name": "axis", "type": "list", "limits": stage_names},
            ],},
    ] + comon_parameters
    ##########################################################

    def __init__(self, parent=None, params_state=None):
        super().__init__(parent, params_state)

        self.controller = None
        self.settings.child("epsilon").setValue(0.002)

    def commit_settings(self, param):
        if param.name() == 'amplitude':
            self.controller.amplitude = Q_(param.value(), units='V')
        elif param.name() == 'frequency':
            self.controller.frequency = Q_(param.value(), 'Hz')

        elif param.name() == 'maxfreq':
            self.controller.max_frequency = Q_(param.value(), 'Hz')
            self.settings.child('maxfreq').setValue(self.controller.max_frequency.m_as('Hz'))

    def ini_stage(self, controller=None):
        """Initialize the controller and stages (axes) with given parameters.

        Parameters
        ----------
        controller (object): custom object of a PyMoDAQ plugin (Slave case).
            None if only one actuator by controller (Master case)

        Returns
        -------
        self.status: (edict) with initialization status: three fields:
            * info: (str)
            * controller: (object) initialized controller
            * initialized: (bool) False if initialization failed otherwise
                True
        """
        try:

            self.status.update(edict(
                info="", controller=None, initialized=False))

            # check whether this stage is controlled by a multiaxe controller
            # (to be defined for each plugin)
            # if multiaxes then init the controller here if Master state
            # otherwise use external controller
            if self.settings.child('multiaxes', 'ismultiaxes').value() \
                    and self.settings.child('multiaxes', 'multi_status').value() == "Slave":
                if controller is None:
                    raise Exception('No controller has been defined externally'
                                    ' while this axe is a slave one')
                else:
                    self.controller = controller
            else:  # Master stage
                index = self.settings.child('device').opts['limits'].index(self.settings.child('device').value())
                self.controller = instrument(psets[index])
                self.settings.child('units').setValue(self.controller.units)
                self.settings.child('amplitude').setOpts(limits=list(self.controller.amplitude_limits.m_as('V')))
                self.settings.child('frequency').setOpts(limits=list(self.controller.frequency_limits.m_as('Hz')))
                self.settings.child('amplitude').setValue(self.controller.amplitude.m_as('V'))
                self.settings.child('frequency').setValue(self.controller.frequency.m_as('Hz'))
                self.settings.child('maxfreq').setValue(self.controller.max_frequency.m_as('Hz'))

            self.status.controller = self.controller
            self.status.initialized = True

            return self.status

        except Exception as e:
            self.emit_status(ThreadCommand("Update_Status", [str(e), "log"]))
            self.status.info = str(e)
            self.status.initialized = False

            return self.status

    def close(self):
        """Close the communication with the SmarAct controller.
        """

        self.controller.close()
        self.controller = None

    def check_position(self):
        """Get the current position from the hardware with scaling conversion.

        Returns
        -------
        float: The position obtained after scaling conversion.
        """
        position = self.controller.check_position().magnitude
        # convert position if scaling options have been used, mandatory here
        position = self.get_position_with_scaling(position)
        #position = self.target_position
        self.current_position = position

        self.emit_status(ThreadCommand("check_position", [position]))

        return position

    def move_Abs(self, position):
        """Move to an absolute position

        Parameters
        ----------
        position: float
        """
        # limit position if bounds options has been selected and if position is
        # out of them
        position = self.check_bound(position)
        self.target_position = position
        # convert the user set position to the controller position if scaling
        # has been activated by user
        position = self.set_position_with_scaling(position)

        self.controller.move_to(Q_(position, self.settings['units']), 'abs')


    def move_Rel(self, position):
        """Move to a relative position

        Parameters
        ----------
        position: float
        """
        position = (self.check_bound(self.current_position + position) - self.current_position)
        self.target_position = position + self.current_position
        position = self.set_position_relative_with_scaling(position)

        self.controller.move_to(Q_(position, self.settings['units']), 'rel')


    def move_Home(self):
        """Move to home and reset position to zero.
        """
        self.controller.move_home()
        self.check_position()

    def stop_motion(self):
        """
        """
        self.controller.stop()
        self.move_done()


if __name__ == "__main__":
    main(__file__)



from pymodaq.daq_move.utility_classes import DAQ_Move_base  # base class
from pymodaq.daq_move.utility_classes import comon_parameters  # common set of
# parameters for all actuators
from pymodaq.daq_utils.daq_utils import ThreadCommand, getLineInfo  # object
# used to send info back to the main thread
from easydict import EasyDict as edict  # type of dict
from ..hardware.smaract.smaract_MCS2_wrapper import SmarActMCS2Wrapper
from ..hardware.smaract.smaract_MCS2_wrapper import get_controller_locators

"""This plugin handles SmarAct MCS2 controller with LINEAR positioners with the
    S option (which means that an encoder is present and give a feedback on the
    current position).
    If you use the first version of MCS controller use the daq_move_SmarActMCS
    plugin.
    The SmarAct MCS2 installer should be executed for this plugin to work.
    We suppose that the configuration of the controller and the positioners
    (sensor type…) has been done via the SmarAct MCS2ServiceTool software.

    If the controller is not switched on, the plugin will not be suggested in
    the list in the GUI of the daq_move.

    It has been tested on Windows 10, with SLC positioner type with enabled
    sensors.
"""


class DAQ_Move_SmarActMCS2(DAQ_Move_base):
    """
        =============== ==============
        **Attributes**    **Type**
        *params*          dictionnary
        =============== ==============
    """
    _controller_units = 'µm'
    controller_locators = get_controller_locators()
    is_multiaxes = True  # we suppose a have a MCS2 controller with a sensor
    # module for 3 channels (stages).
    stage_names = [0, 1, 2]  # be careful that the channel index starts at 0
    # and not at 1 has is done in MCS2ServiceTool
    min_bound = -61500  # µm
    max_bound = +61500  # µm
    # bounds corresponding to the SLC-24180
    offset = 0  # µm

    params = [
                 {'title': 'group parameter:',
                  'name': 'group_parameter',
                  'type': 'group',
                  'children': [
                     {'title': 'Controller Name:',
                      'name': 'smaract_mcs2',
                      'type': 'str',
                      'value': 'SmarAct MCS2 controller',
                      'readonly': True},
                     {'title': 'Controller locator',
                      'name': 'controller_locator', 'type': 'list',
                      'limits': controller_locators},
                  ]},
                 # parameters not specific to the plugin ######################
                 {'title': 'MultiAxes:',
                  'name': 'multiaxes',
                  'type': 'group',
                  'visible': is_multiaxes,
                  'children': [
                     {'title': 'is Multiaxes:',
                      'name': 'ismultiaxes',
                      'type': 'bool',
                      'value': is_multiaxes,
                      'default': False},
                     {'title': 'Status:',
                      'name': 'multi_status',
                      'type': 'list',
                      'value': 'Master',
                      'limits': ['Master', 'Slave']},
                     {'title': 'Axis:',
                      'name': 'axis',
                      'type': 'list',
                      'limits': stage_names},
                  ]}
                 ##############################################################
             ] + comon_parameters

    def __init__(self, parent=None, params_state=None):
        """Initialize the the class

        Parameters
        ----------
        parent (caller object of this plugin): see DAQ_Move_main.DAQ_Move_stage
        params_state (list of dicts): saved state of the plugin parameters list
        """
        super().__init__(parent, params_state)
        self.controller = None
        self.settings.child('epsilon').setValue(0.005)  # this means that we
        # tolerate an error of 5 nanometers on the target position

    def check_position(self):
        """Get the current position from the hardware with scaling conversion.

        Returns
        -------
        float: The position obtained after scaling conversion.
        """
        position = self.controller.get_position(
            self.settings.child('multiaxes', 'axis').value())

        position = float(position) / 1e6  # the position given by the
        # controller is in picometers, we convert in micrometers
        position = self.get_position_with_scaling(position)  # convert position
        # if scaling options have been used, mandatory here

        self.current_position = position

        self.emit_status(ThreadCommand('check_position', [position]))
        return position

    def close(self):
        """Close the communication with the MCS2 controller.
        """
        self.controller.close_communication()
        self.controller = None

    def commit_settings(self, param):
        """
            | Activate any parameter changes on the hardware.
            |
            | Called after a param_tree_changed signal from DAQ_Move_main.

        ## TODO for your custom plugin
        if param.name() == "a_parameter_you've_added_in_self.params":
           self.controller.your_method_to_apply_this_param_change()
        elif ...
        ##
        """

    def ini_stage(self, controller=None):
        """Initialize the controller and stages (axes) with given parameters.

        Parameters
        ----------
        controller (object): custom object of a PyMoDAQ plugin (Slave case).
            None if only one actuator by controller (Master case)

        Returns
        -------
        self.status: (edict) with initialization status: three fields:
            * info: (str)
            * controller: (object) initialized controller
            * initialized: (bool) False if initialization failed otherwise
                True
        """
        try:
            # initialize the stage and its controller status
            # controller is an object that may be passed to other instances of
            # DAQ_Move_Mock in case of one controller controlling
            # multiactuators (or detector)

            self.status.update(edict(
                info="", controller=None, initialized=False
            ))

            # check whether this stage is controlled by a multiaxe controller
            # (to be defined for each plugin)
            # if multiaxes then init the controller here if Master state
            # otherwise use external controller
            if self.settings.child('multiaxes',
                                   'ismultiaxes').value() \
                    and self.settings.child('multiaxes',
                                            'multi_status').value() == "Slave":
                if controller is None:
                    raise Exception('No controller has been defined externally'
                                    ' while this axe is a slave one')
                else:
                    self.controller = controller
            else:  # Master stage
                self.controller = SmarActMCS2Wrapper()
                self.controller.init_communication(
                    self.settings.child('group_parameter',
                                        'controller_locator').value())

            # min and max bounds will depend on which positionner is plugged.
            # Anyway the bounds are secured by the library functions.
            self.settings.child('bounds', 'is_bounds').setValue(True)
            self.settings.child('bounds', 'min_bound').setValue(self.min_bound)
            self.settings.child('bounds', 'max_bound').setValue(self.max_bound)
            self.settings.child('scaling', 'use_scaling').setValue(True)
            self.settings.child('scaling', 'offset').setValue(self.offset)

            self.status.info = "Stage initialized !"
            self.status.controller = self.controller
            self.status.initialized = True
            return self.status

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status',
                                           [getLineInfo() + str(e), 'log']))
            self.status.info = getLineInfo() + str(e)
            self.status.initialized = False
            return self.status

    def move_Abs(self, position):
        """ Move the actuator to the absolute target defined by position

        Parameters
        ----------
        position: (float) value of the absolute target positioning in
            micrometers
        """
        # if the user checked bounds, they are applied here
        position = self.check_bound(position)
        # apply scaling if the user specified one
        position = self.set_position_with_scaling(position)
        self.target_position = position  # in micrometers

        # convert position in picometers
        position = int(position*1e6)
        self.controller.absolute_move(
            self.settings.child('multiaxes', 'axis').value(), position)

        # start a loop to poll the current actuator value and compare it with
        # target position
        self.poll_moving()

    def move_Rel(self, relative_move):
        """ Move the actuator to the relative target value defined by
        relative_move

        Parameters
        ----------
        relative_move: (float) value of the relative distance to travel in
            micrometers
        """
        absolute_position = self.check_bound(self.current_position
                                             + relative_move)
        self.target_position = absolute_position  # in micrometers

        # convert relative_move in picometers
        relative_move = int(relative_move*1e6)
        self.controller.relative_move(
            self.settings.child('multiaxes', 'axis').value(),
            relative_move)

        self.poll_moving()

    def move_Home(self):
        """Move to the physical reference and reset position to 0
        """
        self.controller.find_reference(
            self.settings.child('multiaxes', 'axis').value())
        self.emit_status(ThreadCommand('Update_Status',
                                       ['The positioner has been referenced']))

    def stop_motion(self):
        """Stop the current motion even if the target has not been reached
        """
        self.controller.stop(self.settings.child('multiaxes', 'axis').value())
        self.emit_status(ThreadCommand('Update_Status',
                                       ['The positioner has been stopped']))
        self.move_done()  # to let the interface know the actuator stopped
